from __future__ import absolute_import

from pkg_resources import resource_filename

from io import StringIO
import os
import re
import psi4
import numpy as np

#: convert bohr to angstrom
BOHR_TO_ANGSTROM = 0.52917721092


def molfile(file):
    path = os.path.join('molecules', file)
    return resource_filename(__name__, path)


def chargefile(file):
    path = os.path.join('charges', file)
    return resource_filename(__name__, path)


def datafile(file):
    path = os.path.join('data', file)
    return resource_filename(__name__, path)


def mol_from_file(file):
    with open(molfile(file), 'r') as f:
        geom = f.read()
    mol = psi4.core.Molecule.from_string(geom, fix_com=True,
                                         fix_orientation=True)
    mol.update_geometry()
    return mol


def mol_from_mol2(file):
    with open(molfile(file), 'r') as f:
        atoms = f.read().split('@<TRIPOS>ATOM\n')[1]
        atoms = atoms.split('\n@<TRIPOS>BOND')[0]
    arr = np.loadtxt(StringIO(atoms), usecols=(1, 2, 3, 4), dtype=object)
    arr[:, 0] = [re.sub(r'\d+', '', x) for x in arr[:, 0]]
    rows = [' '.join(row) for row in arr]
    xyz = '{}\n\n{}'.format(len(arr), '\n'.join(rows))
    mol = psi4.core.Molecule.from_string(xyz, fix_com=True,
                                         fix_orientation=True)
    mol.update_geometry()
    return mol


def charges_from_mol2(file):
    with open(molfile(file), 'r') as f:
        atoms = f.read().split('@<TRIPOS>ATOM\n')[1]
        atoms = atoms.split('\n@<TRIPOS>BOND')[0]
    arr = np.loadtxt(StringIO(atoms), usecols=8)
    return arr


def coordinates_from_xyz(file):
    return np.loadtxt(molfile(file), skiprows=2, usecols=(1, 2, 3),
                      comments='!')


def charges_from_red_file(file):
    with open(chargefile(file), 'r') as f:
        content = f.read()

    mols = [x.split('\n')[1:] for x in content.split('MOLECULE') if x]
    charges = [np.array([float(x.split()[4]) for x in y if x]) for y in mols]
    if len(charges) == 1:
        charges = charges[0]
    return charges


def charges_from_itp_file(file):
    with open(chargefile(file), 'r') as f:
        content = f.read()
    section = content.split('[ atoms ]')[1].split('; total charge')[0]
    lines = [x.strip() for x in section.split('\n')]
    lines = [x for x in lines if x and not x.startswith(';')]
    return np.array([float(x.split()[6]) for x in lines])


def grid_from_file(file):
    return np.loadtxt(datafile(file), usecols=(1, 2, 3), comments='!')


def esp_from_gamess_file(file):
    bohr = np.loadtxt(datafile(file), comments='!')
    bohr[:, 1:] *= BOHR_TO_ANGSTROM
    return bohr



"""
Unit and regression test for the psiresp package.
"""

# Import package, test suite, and other packages as needed
import psiresp
import pytest
import sys
import numpy as np

from numpy.testing import assert_almost_equal, assert_allclose
from .utils import mol_from_file, charges_from_red_file, datafile

@pytest.mark.parametrize('stage_2,a,redname', [
    (False, 0.01, 'respA2'),  # really off
    (True, 0.0005, 'respA1'),
    (False, 0.0, 'espA1')
])
class TestMultiRespNoOptNoOrient(object):
    opt = False
    load_files = False
    nme2ala2_names = ['nme2ala2_opt_c1.xyz', 'nme2ala2_opt_c2.xyz']
    methylammonium_names = ['methylammonium_opt_c1.xyz']
    n_orient = 0
    orient = [[(1, 5, 7), (7, 5, 1)],
              [(5, 18, 19), (19, 18, 5), (6, 19, 20), (20, 19, 6)]]

    intra_chrconstr = [[], {0: [20, 21, 22, 23, 24, 25],
                            0.6163: [18],
                            -0.5722: [19]}]
    intra_chrequiv = [[[6, 7, 8]], [[10, 14], [11, 12, 13, 15, 16, 17]]]
    inter_chrconstr = {0: [(1, [1, 2, 3, 4]), (2, [1, 2, 3, 4, 5, 6, 7, 8])]}

    GRID = datafile('test_multiresp/grid.dat')
    ESP = datafile('test_multiresp/grid_esp.dat')
    rtol = 0.01
    atol = 1e-4

    @pytest.fixture()
    def nme2ala2(self):
        mols = [mol_from_file(f) for f in self.nme2ala2_names]
        resp = psiresp.Resp.from_molecules(mols, charge=0,
                                           orient=self.orient[1],
                                           name='nme2ala2',
                                           load_files=self.load_files,
                                           grid_name=self.GRID,
                                           esp_name=self.ESP)
        return resp

    @pytest.fixture()
    def methylammonium(self):
        mols = [mol_from_file(f) for f in self.methylammonium_names]
        resp = psiresp.Resp.from_molecules(mols, charge=1,
                                           orient=self.orient[0],
                                           name='methylammonium',
                                           load_files=self.load_files,
                                           grid_name=self.GRID,
                                           esp_name=self.ESP)
        return resp

    @pytest.fixture()
    def nme2ala2_charges(self, redname):
        fn = 'nme2ala2_multifit_constr_c2_o4_{}.dat'.format(redname)
        return charges_from_red_file(fn)

    @pytest.fixture()
    def multifit_charges(self, redname):
        fn = 'amm_dimethyl_{}.dat'.format(redname)
        return charges_from_red_file(fn)

    def test_single_mol_in_multiresp(self, stage_2, a, nme2ala2,
                                     nme2ala2_charges, tmpdir):
        r = psiresp.MultiResp([nme2ala2])
        with tmpdir.as_cwd():
            charges = r.run(stage_2=stage_2, opt=self.opt, hyp_a1=a,
                            intra_chrequiv=self.intra_chrequiv[1:],
                            intra_chrconstr=self.intra_chrconstr[1:],
                            n_orient=self.n_orient, save_files=False)
        assert_allclose(charges[0], nme2ala2_charges, rtol=self.rtol, 
                        atol=self.atol)

    def test_multi_mol(self, stage_2, a, nme2ala2, methylammonium,
                       multifit_charges, tmpdir):
        r = psiresp.MultiResp([methylammonium, nme2ala2])
        with tmpdir.as_cwd():
            charges = r.run(stage_2=stage_2, opt=self.opt, hyp_a1=a,
                            inter_chrconstr=self.inter_chrconstr,
                            intra_chrequiv=self.intra_chrequiv,
                            intra_chrconstr=self.intra_chrconstr,
                            n_orient=self.n_orient, save_files=False)
        for charge, ref in zip(charges, multifit_charges):
            assert_allclose(charge, ref, rtol=self.rtol, atol=self.atol)



class TestLoadMultiResp(TestMultiRespNoOptNoOrient):
    load_files = True

@pytest.mark.skip(reason='fails even with these looser comparison constraints. orientations are important!')
class TestMultiRespNoOptAutoOrient(TestMultiRespNoOptNoOrient):
    n_orient = 4
    orient = [[], []]
    rtol = 0.15  # will have different orientations
    atol = 1e-2


@pytest.mark.optimize
@pytest.mark.slow
class TestMultiRespOptNoOrient(TestMultiRespNoOptNoOrient):
    opt = True
    nme2ala2_names = ['nme2ala2_c1.xyz', 'nme2ala2_c2.xyz']
    methylammonium_names = ['methylammonium_c1.xyz']



import pytest
import os
import numpy as np

from psiresp import utils
from numpy.testing import assert_almost_equal
from .utils import coordinates_from_xyz, datafile


@pytest.mark.fast
@pytest.mark.parametrize('molname,onum,orient', [
    ('dmso_opt_c1', 1, (1, 5, 6)),
    ('dmso_opt_c1', 2, (6, 5, 1)),
])
def test_orient_rigid(molname, onum, orient):
    original = coordinates_from_xyz('{}.xyz'.format(molname))
    i, j, k = np.asarray(orient)-1
    oriented = utils.orient_rigid(i, j, k, original)
    ref = coordinates_from_xyz('{}_o{}.xyz'.format(molname, onum))
    assert_almost_equal(oriented, ref, decimal=5)


@pytest.mark.fast
@pytest.mark.parametrize('molname,onum,rotate', [
    ('dmso_opt_c1', 3, (1, 5, 6)),
    ('dmso_opt_c1', 4, (6, 5, 1)),
])
def test_rotate_rigid(molname, onum, rotate):
    original = coordinates_from_xyz('{}_qmra.xyz'.format(molname))
    i, j, k = np.asarray(rotate)-1
    oriented = utils.rotate_rigid(i, j, k, original)
    ref = coordinates_from_xyz('{}_o{}.xyz'.format(molname, onum))
    assert_almost_equal(oriented, ref, decimal=5)


@pytest.mark.fast
@pytest.mark.parametrize('factor,radii_name,scaled', [
    (1.4, 'msk', [1.68, 2.1, 1.96]),
    (1.6, 'msk', [1.92, 2.4, 2.24]),
    (1.8, 'msk', [2.16, 2.7, 2.52]),
    (2.0, 'msk', [2.4, 3., 2.8]),
    (1.4, 'bondi', [1.68, 2.38, 2.128]),
    (1.6, 'bondi', [1.92, 2.72, 2.432]),
    (1.8, 'bondi', [2.16, 3.06, 2.736]),
    (2.0, 'bondi', [2.4, 3.4, 3.04])
])
def test_scale_radii(factor, radii_name, scaled):
    radii = utils.scale_radii({'H', 'C', 'O'}, factor,
                              use_radii=radii_name)
    ref = {k: v for k, v in zip('HCO', scaled)}
    assert len(ref) == len(radii)
    for k, refv in ref.items():
        assert_almost_equal(radii[k], refv)


@pytest.mark.fast
@pytest.mark.parametrize('n', [3, 10, 29, 44, 48, 64])
def test_gen_unit_sphere(n):
    points = utils.gen_unit_sphere(n)
    fn = datafile('surface_n{}.dat'.format(n))
    ref = np.loadtxt(fn, comments='!')
    assert_almost_equal(points, ref, decimal=5)
    assert len(points) <= n


@pytest.mark.fast
@pytest.mark.parametrize('scaled,density,n_points', [
    ([1.68, 2.1, 1.96], 0.0, [0, 0, 0]),
    ([1.68, 2.1, 1.96], 1.0, [35, 55, 48]),
    ([1.68, 2.1, 1.96], 2.0, [70, 110, 96]),
    ([2.4, 3., 2.8], 3.4, [246, 384, 334]),
])
def test_gen_connolly_spheres(scaled, density, n_points):
    radii = {k: v for k, v in zip('HCO', scaled)}
    points, rad = utils.gen_connolly_spheres('CHOHCC', radii,
                                             density=density)
    assert len(rad) == 6
    assert len(points) == 6

    point_arr = np.asarray(n_points)[[1, 0, 2, 0, 1, 1]]
    for pt, n in zip(points, point_arr):
        assert len(pt) <= n and len(pt) >= 0.85*n


@pytest.mark.fast
@pytest.mark.parametrize('vdw_radii,scale_factors,density,radii,n_points', [
    ({}, (1.4, 2.0), 1.0,
     [[1.68, 2.1, 1.96], [2.4, 3., 2.8]],
     [[35, 55, 48], [72, 113, 98]]),
    ({}, (1.4, 2.0), 2.0,
     [[1.68, 2.1, 1.96], [2.4, 3., 2.8]],
     [[70, 110, 96], [144, 226, 197]]),
    ({'C': 1}, (2.0,), 2.0, [[2.4, 2, 2.8]], [[144, 100, 197]]),
])
def test_gen_connolly_shells(vdw_radii, scale_factors, density, radii,
                             n_points):
    points = utils.gen_connolly_shells('CHOHCC', vdw_radii=vdw_radii,
                                       scale_factors=scale_factors,
                                       density=density)
    assert len(points) == len(scale_factors)
    assert all(len(spheres[0]) == 6 for spheres in points)
    for (spheres, rads), ns, scaled in zip(points, n_points, radii):
        point_arr = np.asarray(ns)[[1, 0, 2, 0, 1, 1]]
        rad_arr = np.asarray(scaled)[[1, 0, 2, 0, 1, 1]]
        assert_almost_equal(rads, rad_arr)
        for pt, n in zip(spheres, point_arr):
            assert len(pt) <= n and len(pt) >= 0.85*n

# TODO: test_gen_vdw_surface




import psiresp
import pytest
import os
import numpy as np

from numpy.testing import (assert_almost_equal, assert_equal,
                           assert_allclose)
from .utils import mol_from_file


class TestConformer:

    @pytest.fixture()
    def geometry(self):
        return mol_from_file('nme2ala2_c1.xyz')

    @pytest.fixture()
    def opt_mol(self):
        return mol_from_file('nme2ala2_opt_c1.xyz')

    @pytest.fixture(scope='function')
    def conformer(self, geometry):
        return psiresp.Conformer(geometry)

    @pytest.mark.fast
    def test_init_conformer_defaults(self, conformer):
        assert conformer.name == 'default'
        assert conformer.charge == 0
        assert conformer.molecule.molecular_charge() == 0
        assert conformer.multiplicity == 1
        assert conformer.molecule.multiplicity() == 1
        assert conformer.n_atoms == 25
        assert_equal(conformer.symbols, list('CHHHCONHCCHHHCHHHCONHCHHH'))
        assert len(conformer.orientations) == 1
        assert len(conformer._orientations) == 0

    @pytest.mark.fast
    def test_init_conformer_options(self, geometry):
        orient = [(5, 18, 19), (19, 18, 5)]
        conformer = psiresp.Conformer(geometry, name='nme2ala2', charge=2,
                                      multiplicity=2, orient=orient)
        assert conformer.name == 'nme2ala2'
        assert conformer.charge == 2
        assert conformer.molecule.molecular_charge() == 2
        assert conformer.multiplicity == 2
        assert conformer.molecule.multiplicity() == 2
        assert conformer.n_atoms == 25
        assert_equal(conformer.symbols, list('CHHHCONHCCHHHCHHHCONHCHHH'))
        assert len(conformer._orientations) == 0
        assert len(conformer.orientations) == 2
        assert len(conformer._orientations) == 2
        assert conformer.orientations[0].name == 'nme2ala2_o1'
        assert conformer.orientations[1].name == 'nme2ala2_o2'

    @pytest.mark.fast
    def test_add_orientations(self, conformer):
        assert len(conformer.orientations) == 1  # original molecule
        conformer.add_orientations(orient=[(5, 18, 19), (19, 18, 5)])
        assert len(conformer.orientations) == 2
        conformer.add_orientations(orient=[(6, 19, 20), (20, 19, 6)])
        assert len(conformer.orientations) == 4

    @pytest.mark.optimize
    @pytest.mark.slow
    def test_add_orientations_opt(self):
        conformer = psiresp.Conformer(mol_from_file('nme2ala2_c1.xyz'))
        assert len(conformer.orientations) == 1  # original molecule
        conformer.add_orientations(orient=[(5, 18, 19), (19, 18, 5)])
        xyz_pre_opt = conformer.orientations[0].coordinates
        conformer.optimize_geometry()
        xyz_post_opt = conformer.orientations[0].coordinates
        diff = np.linalg.norm(xyz_post_opt-xyz_pre_opt)
        assert diff > 0.01

    @pytest.mark.optimize
    @pytest.mark.slow
    @pytest.mark.parametrize('save_xyz,save_files', [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ])
    def test_optimize_geometry(self, conformer, opt_mol, save_xyz,
                               save_files, tmpdir):
        xyz = 'default_opt.xyz'
        log = 'default_opt.log'
        opt = opt_mol.geometry().np
        with tmpdir.as_cwd():
            conformer.optimize_geometry(save_opt_geometry=save_xyz,
                                        save_files=save_files)
            assert_allclose(conformer.molecule.geometry().np, opt,
                            rtol=0.05, atol=1e-4)
            if save_files:
                assert os.path.exists(log)
            else:
                assert not os.path.exists(log)

            if save_xyz:
                assert os.path.exists(xyz)
            else:
                assert not os.path.exists(xyz)

    @pytest.mark.fast
    def test_clone(self, conformer):
        new = conformer.clone()
        assert new.molecule is not conformer.molecule
        assert_almost_equal(new.molecule.geometry().np,
                            conformer.molecule.geometry().np)
        assert new.name == 'default_copy'
        assert new.charge == 0
        assert new.multiplicity == 1
        assert len(new.orientations) == len(conformer.orientations)
        assert_equal(new._orient, conformer._orient)



"""
Unit and regression test for the psiresp package.
"""

# Import package, test suite, and other packages as needed
import psiresp
import pytest
import sys
import numpy as np

from numpy.testing import assert_almost_equal, assert_allclose
from .utils import (mol_from_file, mol_from_mol2, charges_from_mol2,
                    charges_from_red_file, charges_from_itp_file,
                    datafile)


@pytest.fixture()
def dmso():
    return mol_from_file('dmso_c1.xyz')


@pytest.fixture()
def dmso_opt():
    return mol_from_file('dmso_opt_c1.xyz')


@pytest.fixture()
def dmso_orients():
    fns = ['dmso_opt_c1_o1.xyz', 'dmso_opt_c1_o2.xyz']
    return [mol_from_file(f) for f in fns]


class BaseTestRespConfigNoOpt:
    """Charges from R.E.D. jobs"""

    def test_resp_noopt_noorient(self, dmso_orients, tmpdir, ref):
        r = self.cls.from_molecules(dmso_orients, charge=0)
        r0 = psiresp.Resp.from_molecules(dmso_orients, charge=0)
        with tmpdir.as_cwd():
            charges = r.run(opt=False, equal_methyls=True)
        assert_allclose(charges, ref, rtol=0.05, atol=1e-4)

    def test_resp_noopt_orient(self, dmso_opt, tmpdir, ref):
        r = self.cls.from_molecules([dmso_opt])
        with tmpdir.as_cwd():
            charges = r.run(n_orient=2, opt=False, equal_methyls=True)
        assert_allclose(charges, ref, rtol=0.05, atol=1e-4)


class BaseTestRespConfigOpt:
    """Charges from R.E.D. jobs"""
    name = None
    cls = None
    nconf = 1
    norient = 2

    @property
    def chargefile(self):
        return 'dmso_c{}_o{}_{}.dat'.format(self.nconf, self.norient,
                                            self.name)

    @pytest.fixture()
    def ref(self):
        return charges_from_red_file(self.chargefile)

    @pytest.mark.slow
    def test_resp_opt(self, dmso, tmpdir, ref):
        r = self.cls.from_molecules([dmso])
        with tmpdir.as_cwd():
            charges = r.run(opt=True, n_orient=2, equal_methyls=True,
                            save_files=False, save_opt_geometry=False)
        assert_allclose(charges, ref, rtol=0.05, atol=1e-4)


class TestRespA1(BaseTestRespConfigNoOpt, BaseTestRespConfigOpt):
    cls = psiresp.RespA1
    name = 'respA1'


class TestRespA2(BaseTestRespConfigNoOpt, BaseTestRespConfigOpt):
    cls = psiresp.RespA2
    name = 'respA2'


class TestEspA1(BaseTestRespConfigNoOpt, BaseTestRespConfigOpt):
    cls = psiresp.EspA1
    name = 'espA1'

# whyyyy 😭
@pytest.mark.skip(reason='Psi4 minimises to a very different geometry to GAMESS')
class TestEspA2(BaseTestRespConfigOpt):
    cls = psiresp.EspA2
    name = 'espA2'


class BaseTestATBResp:
    """charges from ATB jobs"""
    cls = psiresp.ATBResp
    name = 'ATB'
    molname = None

    @property
    def chargefile(self):
        return '{}_c1_o0_ATB.itp'.format(self.molname)

    @property
    def molfile(self):
        return '{}_c1_ATB.xyz'.format(self.molname)

    @pytest.fixture(scope='function')
    def resp(self):
        opt = [mol_from_file('{}_c1_ATB_opt.xyz'.format(self.molname))]
        r = self.cls.from_molecules(opt)
        return r

    @pytest.fixture()
    def ref(self):
        return charges_from_itp_file(self.chargefile)

    def test_resp_noopt(self, tmpdir, resp, ref):
        with tmpdir.as_cwd():
            charges = resp.run(opt=False, n_orient=2, equal_methyls=True,
                               vdw_point_density=1, solvent='water')
        # no idea which point density ATB uses
        assert_allclose(charges, ref, rtol=0.05, atol=1e-3)

    @pytest.mark.slow
    def test_resp_opt(self, tmpdir, ref):
        mol = [mol_from_file(self.molfile)]
        r = self.cls.from_molecules(mol)
        with tmpdir.as_cwd():
            charges = r.run(opt=True, n_orient=2, equal_methyls=True,
                            vdw_point_density=1, solvent='water')
        assert_allclose(charges, ref, rtol=0.05, atol=1e-3)


class TestATBRespMethane(BaseTestATBResp):
    molname = 'methane'

# fails, idk why; ATB charges are also different from Malde et al 2011
@pytest.mark.skip(reason='Fails? ATB charges are also v. different from Malde et al 2011')
class TestATBRespIsopropanol(BaseTestATBResp):
    molname = 'isopropanol'


class TestATBRespDMSO(BaseTestATBResp):
    molname = 'dmso'


@pytest.mark.skip(reason='Fails? Are the mol2 geometries not minimised?')
@pytest.mark.resp2
@pytest.mark.slow
@pytest.mark.parametrize('name', ['C00', 'C64', 'MPE', 'PXY', 'TOL'])
@pytest.mark.parametrize('delta', [0, 0.6, 1])
class TestResp2Charges:
    """Charges from RESP2 paper Schauperl et al., 2020.

    See dataset for more: https://doi.org/10.5281/zenodo.3593762
    """

    def test_resp2_noopt(self, name, delta, tmpdir):
        fn = '{}_R2_{:d}.mol2'.format(name, int(delta*100))
        mols = [mol_from_mol2(fn)]
        ref = charges_from_mol2(fn)
        with tmpdir.as_cwd():
            r = psiresp.Resp2.from_molecules(mols, charge=0)
            charges = r.run(opt=False, n_orient=4, delta=delta)
        assert_almost_equal(charges, ref, decimal=3)


class BaseTestResp2Ethanol:
    """Charges from ethanol example in MSchauperl/resp2.

    Assert almost equal with decimal=3 because not sure when values are rounded.
    See repo for more: https://github.com/MSchauperl/RESP2
    """

    solv = np.array([-0.2416,  0.3544, -0.6898,  0.0649,  0.0649,
                     0.0649, -0.0111, -0.0111,  0.4045])

    gas = np.array([-0.2300,  0.3063, -0.5658,  0.0621,  0.0621,
                    0.0621, -0.0153, -0.0153,  0.3339])

    ref = np.array([-0.2358,  0.33035, -0.6278,  0.0635,
                    0.0635,  0.0635, -0.0132, -0.0132,  0.3692])

    def test_ethanol_no_opt(self, tmpdir):
        mols = [mol_from_file('ethanol_resp2_opt_c1.xyz'),
                mol_from_file('ethanol_resp2_opt_c2.xyz')]
        r = psiresp.Resp2.from_molecules(mols, charge=0, name='resp2_ethanol',
                                         load_files=self.load_files,
                                         grid_name=datafile('test_resp2/grid.dat'),
                                         esp_name=datafile('test_resp2/grid_esp.dat'))
        with tmpdir.as_cwd():
            charges = r.run(opt=False, n_orient=0, delta=0.5)
        assert_almost_equal(r.gas_charges, self.gas, decimal=3)
        assert_almost_equal(r.solv_charges, self.solv, decimal=3)
        assert_almost_equal(charges, self.ref, decimal=3)


@pytest.mark.fast
class TestLoadResp2Ethanol(BaseTestResp2Ethanol):
    load_files = True


@pytest.mark.resp2
class TestResp2Ethanol(BaseTestResp2Ethanol):
    load_files = False

    @pytest.mark.slow
    def test_ethanol_opt(self, tmpdir):
        mols = [mol_from_file('ethanol_resp2_c1.xyz'),
                mol_from_file('ethanol_resp2_c2.xyz')]
        r = psiresp.Resp2.from_molecules(mols, charge=0)
        with tmpdir.as_cwd():
            charges = r.run(opt=True, n_orient=0, delta=0.5)
        assert_almost_equal(r.gas_charges, self.gas, decimal=3)
        assert_almost_equal(r.solv_charges, self.solv, decimal=3)
        assert_almost_equal(charges, self.ref, decimal=3)



from __future__ import division, absolute_import
import itertools

import numpy as np

from . import vdwradii


def rotate_x(n, coords):
    """
    Rotate coordinates such that the ``n``th coordinate of
    ``coords`` becomes x-axis.

    Adapted from R.E.D. in perl.

    Parameters
    ----------
    n: int
        index of coordinates
    coords: ndarray
        coordinates
    """
    x, y, z = coords[n]
    hypotenuse = np.sqrt(y**2 + z**2)
    adjacent = abs(y)
    angle = np.arccos(adjacent/hypotenuse)

    if z >= 0:
        if y < 0:
            angle = np.pi - angle
    else:
        if y >= 0:
            angle = 2 * np.pi - angle
        else:
            angle = np.pi + angle

    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    ys = coords[:, 1].copy()
    zs = coords[:, 2].copy()
    coords[:, 1] = zs*sin_angle + ys*cos_angle
    coords[:, 2] = zs*cos_angle - ys*sin_angle


def rotate_z(n, coords):
    """
    Rotate coordinates such that the ``n``th coordinate of
    ``coords`` becomes z-axis.

    Adapted from R.E.D. in perl.

    Parameters
    ----------
    n: int
        index of coordinates
    coords: ndarray
        coordinates
    """
    x, y, z = coords[n]
    hypotenuse = np.sqrt(x**2 + y**2)
    adjacent = abs(x)
    angle = np.arccos(adjacent/hypotenuse)

    if y >= 0:
        if x < 0:
            angle = np.pi - angle
    else:
        if x >= 0:
            angle = 2 * np.pi - angle
        else:
            angle = np.pi + angle

    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    xs = coords[:, 0].copy()
    ys = coords[:, 1].copy()
    coords[:, 0] = xs*cos_angle + ys*sin_angle
    coords[:, 1] = ys*cos_angle - xs*sin_angle


def orient_rigid(i, j, k, coords):
    """
    Rigid-body reorientation such that the ``i`` th coordinate
    is the new origin; the ``j` `th coordinate defines the new 
    x-axis; and the ``k`` th coordinate defines the XY plane.

    ``i``, ``j``, and ``k`` should all be different. They are 
    indexed from 0.

    Adapted from R.E.D. in perl.

    Parameters
    ----------
    i: int
        index. Must be different to ``j`` and ``k``
    j: int
        index. Must be different to ``i`` and ``k``
    k: int
        index. Must be different to ``i`` and ``j``
    coords: ndarray
        coordinates

    Returns
    -------
    coordinates: ndarray
        Re-oriented coordinates
    """
    xyz = coords.copy()
    vec = coords[i]
    xyz -= vec
    rotate_x(j, xyz)
    rotate_z(j, xyz)
    rotate_x(k, xyz)
    return xyz


def rotate_rigid(i, j, k, coords):
    """
    Rigid-body rotation such that the ``i`` th and ``j`` th coordinate 
    define a vector parallel to the x-axis; and the ``k`` th coordinate 
    defines a plane parallel to the XY plane.

    ``i`` , ``j`` , and ``k`` should all be different. They are 
    indexed from 0.

    Adapted from R.E.D. in perl.

    Parameters
    ----------
    i: int
        index. Must be different to ``j`` and ``k``
    j: int
        index. Must be different to ``i`` and ``k``
    k: int
        index. Must be different to ``i`` and ``j``
    coords: ndarray
        coordinates

    Returns
    -------
    coordinates: ndarray
        Rotated coordinates
    """
    vec = coords[i].copy()
    xyz = orient_rigid(i, j, k, coords)
    xyz += vec
    return xyz


def scale_radii(symbols, scale_factor, vdw_radii={}, use_radii='msk'):
    """
    Scale van der Waals' radii

    Parameters
    ----------
    symbols: set
        set of element symbols for each atom
    scale_factor: float
    use_radii: str (optional)
        which set of van der Waals' radii to use
    vdw_radii: dict (optional)
        van der Waals' radii. If elements in the molecule are not
        defined in the chosen ``use_radii`` set, they must be given here.

    Returns
    -------
    radii: dict
        scaled radii
    """

    radii = {}
    vdw_set = vdwradii.options[use_radii.lower()]
    for x in symbols:
        try:
            r = vdw_radii.get(x, vdw_set[x])
        except KeyError:
            err = ('{} is not a supported element. Pass in the '
                   'radius in the ``vdw_radii`` dictionary.')
            raise KeyError(err.format(x))
        else:
            radii[x] = r*scale_factor
    return radii


def gen_unit_sphere(n):
    """
    Get coordinates of n points on a unit sphere. 

    Adapted from GAMESS.

    Parameters
    ----------
    n: int
        maximum number of points

    Returns
    -------
    coordinates: np.ndarray
        cartesian coordinates of points
    """
    pi = np.pi
    n_lat = int((pi*n)**0.5)
    n_long = int((n_lat/2))
    fi = np.arange(n_long+1)*pi/n_long
    z, xy = np.cos(fi), np.sin(fi)
    n_horiz = (xy*n_lat+1e-10).astype(int)
    n_horiz = np.where(n_horiz < 1, 1, n_horiz)
    # get actual points
    dots = np.empty((sum(n_horiz), 3))
    dots[:, -1] = np.repeat(z, n_horiz)
    XY = np.repeat(xy, n_horiz)
    fjs = np.concatenate([2*pi*np.arange(j)/j for j in n_horiz])
    dots[:, 0] = np.cos(fjs)*XY
    dots[:, 1] = np.sin(fjs)*XY
    return dots[:n]


def gen_connolly_spheres(symbols, radii, density=1.0):
    """
    Compute Connolly spheres of specified radii and density around each atom.

    Parameters
    ----------
    symbols: iterable
        array of element symbols for each atom
    radii: dict
        scaled radii of elements
    density: float (optional)
        point density

    Returns
    -------
    points: ndarray
        cartesian coordinates of points
    radii: ndarray
        array of radii of each atom
    """
    rad_arr = np.array([radii[z] for z in symbols])
    rads, inv = np.unique(rad_arr, return_inverse=True)
    n_points = ((rads**2)*np.pi*4*density).astype(int)
    points = [gen_unit_sphere(n)*r for n, r in zip(n_points, rads)]
    all_points = [points[i] for i in inv]  # memory?
    return all_points, rad_arr


def gen_connolly_shells(symbols, vdw_radii={}, use_radii='msk',
                        scale_factors=(1.4, 1.6, 1.8, 2.0),
                        density=1.0):
    """
    Generate Connolly shells for each scale factor for each atom.

    Parameters
    ----------
    symbols: iterable
        array of element symbols for each atom
    use_radii: str (optional)
        which set of van der Waals' radii to use
    scale_factors: iterable of floats (optional)
        scale factors
    density: float (optional)
        point density
    vdw_radii: dict (optional)
        van der Waals' radii. If elements in the molecule are not
        defined in the chosen ``use_radii`` set, they must be given here.

    Returns
    -------
    vdw_points: list
    """
    symbol_set = set([x.capitalize() for x in symbols])
    vdw_points = []
    for sf in scale_factors:
        radii = scale_radii(symbol_set, sf, use_radii=use_radii, vdw_radii=vdw_radii)
        shell = gen_connolly_spheres(symbols, radii, density=density)
        vdw_points.append(shell)
    return vdw_points


def gen_vdw_surface(points, radii, coordinates, rmin=0, rmax=-1):
    """
    Compute van der Waals surface in angstrom.

    Parameters
    ----------
    points: ndarray (n_atoms, 3)
        Cartesian coordinates of unit shells for each atom
    radii: ndarray (n_atoms,)
        Scaled van der Waals' radii of each atom
    coordinates: ndarray (n_atoms, 3)
        Cartesian coordinates of each atom
    rmin: float (optional)
        inner boundary of shell to keep grid points from
    rmax: float (optional)
        outer boundary of shell to keep grid points from. If < 0,
        all points are selected.

    Returns
    -------
    surface_points: ndarray  (n_points, 3)
        Cartesian coordinates in angstrom
    """
    if rmax < 0:
        rmax = np.inf
    if rmax < rmin:
        raise ValueError('rmax must be equal to or greater than rmin')

    if len(points) != len(coordinates):
        err = ('Length of ``points`` must match length of ``coordinates``'
               'Generate unit shells for atoms with gen_connolly_shells()')
        raise ValueError(err)
    if len(radii) != len(coordinates):
        err = ('Length of ``radii`` must match length of ``coordinates``'
               'Generate scaled radii for atoms with gen_connolly_shells()')
        raise ValueError(err)

    radii = np.asarray(radii)
    indices = np.arange(radii.shape[0])
    inner = radii*rmin
    inner = np.where(inner < radii, radii, inner)
    outer = radii*rmax

    surface_points = []
    for i, (dots, xyz) in enumerate(zip(points, coordinates)):
        shell = dots+xyz
        mask = indices != i
        a = np.tile(coordinates[mask], (len(dots), 1, 1))
        b = np.tile(shell.reshape(-1, 1, 3), (1, len(radii)-1, 1))
        dist = np.linalg.norm(a-b, axis=2)
        bounds = (dist >= inner[mask]) & (dist <= outer[mask])
        outside = np.all(bounds, axis=1)
        surface_points.extend(shell[outside])
    return np.array(surface_points)


def isiterable(obj):
    """
    Returns ``True`` if ``obj`` is iterable and not a string

    Adapted from MDAnalysis.lib.util.iterable
    """
    if isinstance(obj, str):
        return False
    if hasattr(obj, 'next'):
        return True
    try:
        len(obj)
    except (TypeError, AttributeError):
        return False
    return True


def asiterable(obj):
    """
    Returns ``obj`` in a list if ``obj`` is not iterable
    """
    if not isiterable(obj):
        return [obj]
    return obj


def empty(obj):
    """Returns if ``obj`` is Falsy"""
    if isinstance(obj, np.ndarray):
        return True
    return not obj


def iter_single(obj):
    """Return iterables of ``obj``, treating an empty list as an object"""
    if not isiterable(obj) or empty(obj):
        return itertools.repeat(obj)
    return asiterable(obj)



from __future__ import division, absolute_import
import warnings
import itertools
import logging

import numpy as np

from .conformer import Conformer
from . import utils

log = logging.getLogger(__name__)


class Resp(object):
    """
    Class to manage R/ESP for one molecule of multiple conformers.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    restrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run
        with restraint=True)
    """

    @classmethod
    def from_molecules(cls, molecules, charge=0, multiplicity=1, name=None,
                       orient=[], rotate=[], translate=[],
                       grid_name='grid.dat', esp_name='grid_esp.dat',
                       load_files=False, **kwargs):
        """
        Create Resp class from Psi4 molecules.

        Parameters
        ----------
        molecules: iterable of Psi4 molecules
            conformers of the molecule. They must all have the same atoms
            in the same order.
        charge: int (optional)
            overall charge of the molecule.
        multiplicity: int (optional)
            multiplicity of the molecule
        name: str (optional)
            name of the molecule. This is used to name output files. If not
            given, defaults to 'Mol'.
        orient: list of tuples of ints (optional)
            List of reorientations to add to each conformer.
            Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        rotate: list of tuples of ints (optional)
            List of rotations to add to each conformer.
            Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        translate: list of tuples of floats (optional)
            List of translations to add to each conformer.
            Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        grid_name: str (optional)
            template for grid filename for each Orientation.
        esp_name: str (optional)
            template for ESP filename for each Orientation.
        load_files: bool (optional)
            If ``True``, tries to load data from file for each Orientation.
        **kwargs:
            arguments passed to ``Resp.__init__()``.

        Returns
        -------
        resp: Resp
        """
        molecules = utils.asiterable(molecules)
        if name is not None:
            names = ['{}_c{}'.format(name, i+1) for i in range(len(molecules))]
        else:
            molnames = [m.name() for m in molecules]
            gennames = ['Mol_c{}'.format(i+1) for i in range(len(molecules))]
            # default Psi4 name is 'default'...
            # Hope no one actually wants to call their molecules default!
            names = np.where(molnames == 'default', molnames, gennames)

        conformers = []
        for mol, name in zip(molecules, names):
            conformers.append(Conformer(mol.clone(), name=name, charge=charge,
                                        multiplicity=multiplicity,
                                        orient=orient, rotate=rotate,
                                        translate=translate,
                                        grid_name=grid_name,
                                        esp_name=esp_name,
                                        load_files=load_files))

        return cls(conformers, name=name, **kwargs)

    def __init__(self, conformers, name=None, chrconstr=[], chrequiv=[]):
        if name is None:
            name = 'Resp'
        if not conformers:
            raise ValueError('Resp must be created with at least one conformer')
        self.name = name
        self.conformers = utils.asiterable(conformers)
        self.n_conf = len(self.conformers)
        self.charge = self._conf.charge
        self.symbols = self._conf.symbols
        self.n_atoms = self._conf.n_atoms
        self.indices = np.arange(self.n_atoms)
        self.atom_ids = self.indices+1
        self.heavy_ids = np.where(self.symbols != 'H')[0]+1
        self.h_ids = np.where(self.symbols == 'H')[0]+1
        self._orient_combs = self._gen_orientation_atoms()
        self.chrconstr = []
        self.chrequiv = []
        if chrconstr is not None:
            self.add_charge_constraints(chrconstr)
        if chrequiv is not None:
            self.add_charge_equivalences(chrequiv)

    @property
    def _conf(self):
        return self.conformers[0]

    @property
    def n_structures(self):
        return sum([c.n_orientations for c in self.conformers])

    def clone(self, name=None):
        """Clone into another instance of Resp

        Parameters
        ----------
        name: str (optional)
            If not given, the new Resp instance has '_copy' appended 
            to its name

        Returns
        -------
        resp: Resp
        """
        if name is None:
            name = self.name+'_copy'
        mols = [c.molecule.clone() for c in self.conformers]
        charge = self.charge
        mult = self._conf.multiplicity
        new = type(self).from_molecules(mols, name=name, charge=charge,
                                        multiplicity=mult)
        for nc, mc in zip(new.conformers, self.conformers):
            nc.add_orientations(orient=mc._orient, rotate=mc._rotate,
                                translate=mc._translate)
        return new

    def add_charge_constraint(self, charge, atom_ids):
        """
        Add charge constraint

        Parameters
        ----------
        charge: float
        atom_ids: iterable of ints
            Atom numbers involved in charge constraint, indexed from 1.
            All the atoms specified here combine for a charge of ``charge``. 
        """
        atom_ids = utils.asiterable(atom_ids)
        if not all(isinstance(x, int) for x in atom_ids):
            raise ValueError('atom_ids must be an iterable of integer atom numbers')
        for a in atom_ids:
            if a not in self.atom_ids:
                raise ValueError('Atom number not found: {}'.format(a))
        self.chrconstr.append([charge, np.asarray(atom_ids)])

    def add_charge_constraints(self, chrconstr=[]):
        """Add charge constraints

        Parameters
        ----------
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        """
        if isinstance(chrconstr, dict):
            chrconstr = list(chrconstr.items())
        for q, ids in chrconstr:
            self.add_charge_constraint(q, ids)

    def add_charge_equivalence(self, atom_ids):
        """
        Add constraint for equivalent charges.

        Parameters
        ----------
        atom_ids: iterable of ints
            Atom numbers involved in charge constraint, indexed from 1.
            All the atoms specified here are constrained to the same charge.
        """
        for a in atom_ids:
            if a not in self.atom_ids:
                raise ValueError('Atom number not found: {}'.format(a))
        if len(atom_ids) < 2:
            raise ValueError('Cannot add equivalence constraint for <2 atoms')
        self.chrequiv.append(np.asarray(atom_ids))

    def add_charge_equivalences(self, chrequiv=[]):
        """Add charge equivalence constraints

        Parameters
        ----------
        chrequiv: list (optional)
            lists of atoms with equivalent charges, indexed from 1.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.
        """
        for ids in chrequiv:
            self.add_charge_equivalence(ids)

    def optimize_geometry(self, method='scf', basis='6-31g*',
                          psi4_options={}, save_opt_geometry=False,
                          save_files=False):
        """
        Optimise geometry for all conformers.

        Parameters
        ----------
        basis: str (optional)
            Basis set to optimise geometry
        method: str (optional)
            Method to optimise geometry
        psi4_options: dict (optional)
            additional Psi4 options
        save_opt_geometry: bool (optional)
            if ``True``, saves optimised geometries to XYZ file
        save_files: bool (optional)
            if ``True``, Psi4 files are saved. This does not affect 
            writing optimised geometries to files. 
        """
        for conf in self.conformers:
            conf.optimize_geometry(method=method, basis=basis,
                                   psi4_options=psi4_options,
                                   save_opt_geometry=save_opt_geometry,
                                   save_files=save_files)

    def set_user_constraints(self, chrconstr=[], chrequiv=[]):
        """
        Get A and B matrices for charge constraints.

        Parameters
        ----------
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        chrequiv: list (optional)
            lists of atoms with equivalent charges, indexed from 1.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.

        Returns
        -------
        A: ndarray
        B: ndarray
        """
        if isinstance(chrconstr, dict):
            chrconstr = list(chrconstr.items())
        else:
            chrconstr = list(chrconstr)

        if chrequiv is None:
            chrequiv = []
        else:
            chrequiv = list(chrequiv)

        chrconstr = self.chrconstr + chrconstr
        chrequiv = self.chrequiv + chrequiv

        equiv = [np.asarray(x)-1 for x in chrequiv if len(x) >= 2]
        edges = np.r_[0, np.cumsum([len(x)-1 for x in equiv])].astype(int)
        n_equiv = edges[-1]
        n_constr = len(chrconstr)
        ndim = n_equiv+n_constr+self.n_atoms+1

        A = np.zeros((ndim, ndim))
        B = np.zeros(ndim)

        for i, (q, ids) in enumerate(chrconstr, self.n_atoms+1):
            B[i] = q
            ix = np.asarray(ids)-1
            A[i, ix] = A[ix, i] = 1

        for i, indices in enumerate(equiv):
            x = np.arange(edges[i], edges[i+1])+self.n_atoms+1+n_constr
            A[(x, indices[:-1])] = A[(indices[:-1], x)] = -1
            A[(x, indices[1:])] = A[(indices[1:], x)] = 1

        return A, B

    def iter_solve(self, q, a, b, hyp_a=0.0005, hyp_b=0.1, ihfree=True,
                   tol=1e-6, maxiter=50):
        """
        Fit the charges iteratively, as required for the hyperbola penalty
        function.

        Parameters
        ----------
        q: ndarray
            partial atomic charges
        a: ndarray
            unrestrained matrix A
        b: ndarray
            matrix B
        hyp_a: float (optional)
            scale factor of asymptote limits of hyperbola
        hyp_b: float (optional)
            tightness of hyperbola at its minimum
        ihfree: bool (optional)
            if `True`, exclude hydrogens from restraint
        tol: float (optional)
            threshold for convergence
        maxiter: int (optional)
            maximum number of iterations

        Returns
        -------
        charges: ndarray
        """
        if not hyp_a:  # i.e. no restraint
            return q

        mask = np.ones(self.n_atoms, dtype=bool)
        if ihfree:
            mask[self.h_ids-1] = False
        diag = np.diag_indices(self.n_atoms)
        ix = (diag[0][mask], diag[1][mask])
        indices = np.where(mask)[0]
        b2 = hyp_b**2

        niter, delta = 0, 2*tol
        while delta > tol and niter < maxiter:
            q_last, a_i = q.copy(), a.copy()
            a_i[ix] = a[ix] + hyp_a/np.sqrt(q[indices]**2 + b2) * self.n_structures
            q = np.linalg.solve(a_i, b)
            delta = np.max((q-q_last)[:self.n_atoms]**2) ** 0.5
            niter += 1

        if delta > tol:
            err = 'Charge fitting did not converge with maxiter={}'
            warnings.warn(err.format(maxiter))

        return q

    def get_constraint_matrices(self, chrconstr=[], chrequiv=[],
                                weights=1, use_radii='msk',
                                vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                vdw_point_density=1.0, vdw_radii={},
                                rmin=0, rmax=-1, basis='6-31g*', method='scf',
                                solvent=None, psi4_options={},
                                save_files=False):
        """
        Get A and B matrices to solve for charges, including charge constraints.

        Parameters
        ----------
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        chrequiv: list (optional)
            lists of atoms with equivalent charges, indexed from 1.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.
        weights: iterable (optional)
            weights of each conformer. If only one number is given, this is 
            repeated for each conformer.
        use_radii: str (optional)
            which set of van der Waals' radii to use
        vdw_scale_factors: iterable of floats (optional)
            scale factors
        vdw_point_density: float (optional)
            point density
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        basis: str (optional)
            Basis set to compute ESP
        method: str (optional)
            Method to compute ESP
        solvent: str (optional)
            Solvent for computing in implicit solvent
        psi4_options: dict (optional)
            additional Psi4 options
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed ESP
            and grids are written to files.

        Returns
        -------
        a: ndarray
        b: ndarray
        """

        A, B = self.set_user_constraints(chrconstr=chrconstr,
                                         chrequiv=chrequiv)

        if not utils.isiterable(weights):
            weights = itertools.repeat(weights)
        elif len(weights) != self.n_conf:
            err = ('weights must be an iterable of the same length '
                   'as number of conformers')
            raise ValueError(err)

        # don't compute anything unless we have to
        if any(conf._grid_needs_computing for conf in self.conformers):
            vdw_points = utils.gen_connolly_shells(self.symbols,
                                                   vdw_radii=vdw_radii,
                                                   use_radii=use_radii,
                                                   scale_factors=vdw_scale_factors,
                                                   density=vdw_point_density)
        else:
            vdw_points = []

        # get molecule weights
        n_a = self.n_atoms
        for conf, w in zip(self.conformers, weights):
            a, b = conf.get_esp_matrices(weight=w, vdw_points=vdw_points,
                                         rmin=0, rmax=-1, basis=basis,
                                         method=method, solvent=solvent,
                                         psi4_options=psi4_options,
                                         save_files=save_files)
            A[:n_a, :n_a] += a
            B[:n_a] += b

        A[n_a, :n_a] = A[:n_a, n_a] = 1
        B[n_a] = self.charge

        return A, B

    def fit(self, restraint=True, hyp_a=0.0005, hyp_b=0.1, ihfree=True,
            tol=1e-6, maxiter=50, **kwargs):
        """
        Perform the R/ESP fits.

        Parameters
        ----------
        restraint: bool (optional)
            whether to perform a restrained fit
        hyp_a: float (optional)
            scale factor of asymptote limits of hyperbola
        hyp_b: float (optional)
            tightness of hyperbola at its minimum
        ihfree: bool (optional)
            if `True`, exclude hydrogens from restraint
        tol: float (optional)
            threshold for convergence
        maxiter: int (optional)
            maximum number of iterations
        **kwargs:
            arguments passed to Resp.get_constraint_matrices

        Returns
        -------
        charges: ndarray
        """
        a, b = self.get_constraint_matrices(**kwargs)
        q = np.linalg.solve(a, b)
        self.unrestrained_charges = q[:self.n_atoms]
        if restraint:
            q = self.iter_solve(q, a, b, hyp_a=hyp_a, hyp_b=hyp_b,
                                ihfree=ihfree, tol=tol, maxiter=maxiter)
            self.restrained_charges = q[:self.n_atoms]

        self.charges = q[:self.n_atoms]
        return self.charges

    def get_sp3_ch_ids(self):
        """
        Get atom numbers of hydrogens bonded to sp3 carbons. Numbers
        are indexed from 1.

        Returns
        -------
        groups: dict
            dict of {C atom number: array of bonded H numbers}.
        """
        import psi4

        groups = {}
        bonds = psi4.qcdb.parker._bond_profile(self._conf.molecule)
        bonds = np.asarray(bonds)[:, :2]  # [[i, j, bond_order]]
        for i in self.indices[self.symbols == 'C']:
            cbonds = bonds[np.any(bonds == i, axis=1)]
            partners = np.ravel(cbonds[cbonds != i])
            groups[i+1] = partners[self.symbols[partners] == 'H']+1
        return groups

    def get_methyl_constraints(self, chrconstr=None):
        """
        Get charge equivalence arrays when all methyls are treated as
        equivalent, and all methylenes are equivalent. Toggle this with
        ``equal_methyls=True`` in ``run()``.

        Parameters
        ----------
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.

        Returns
        -------
        equivalence arrays: list of lists of ints
            List of equivalence arrays. First array contains methyl carbons;
            second contains methylene carbons; third contains methyl hydrogens;
            last contains methylene hydrogens.
        """
        cs, hs = zip(*self.get_sp3_ch_ids().items())
        if chrconstr is None:
            flat_constr = []
        else:
            flat_constr = np.ravel(chrconstr)
        equivs = [c3s, c2s, h3s, h2s] = [[], [], [], []]
        for c, h in zip(cs, hs):
            if c in flat_constr or any(x in flat_constr for x in h):
                continue
            if len(h) == 3:
                c3s.append(c)
                h3s.extend(h)
            else:
                c2s.append(c)
                h2s.extend(h)
        return equivs

    def get_stage2_constraints(self, q, equal_methyls=False,
                               chrequiv=[], chrconstr=[]):
        """
        Create constraints for RESP stage 2. Atom numbers are indexed from 1.

        If charge equivalence constraints are provided, these are used in stage
        2 fitting. Otherwise, if ``equal_methyls=True``, the carbons in
        methyl groups are constrained to be equivalent, as are the carbons in
        methylenes, and the hydrogens in them. If ``equal_methyls=False``,
        equivalence constraints are created for each group of hydrogens bonded
        to sp3 carbons, separately.

        Any atoms involved in charge constraints are not included in the
        equivalence constraints.

        Parameters
        ----------
        q: ndarray
            charges from stage 1
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent. Ignored if ``chrequiv`` constraints are provided.
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        chrequiv: list (optional)
            lists of atoms with equivalent charges, indexed from 1.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.

        Returns
        -------
        chrconstr: list of tuples
            Charge constraints in the form of [(charge, atom_number_list)]
        chrequiv: list
            Lists of atoms with equivalent charges in the form
            [atom_number_list]
        """
        cs = []
        if chrequiv:  # constraints are fitted in stage 2
            equivs = chrequiv
        elif equal_methyls:
            equivs = [a, b, c, d] = self.get_methyl_constraints(chrconstr)
        else:
            cs, equivs = zip(*self.get_sp3_ch_ids().items())
        chs = np.r_[cs, np.concatenate(equivs)]

        q = np.asarray(q)
        ids = self.indices+1
        mask = ~np.in1d(ids, chs)
        constraints = [(q, [a]) for q, a in zip(q[mask], ids[mask])]
        return constraints, equivs

    def _gen_orientation_atoms(self):
        """
        Generate potential orientations from heavy atoms first,
        then hydrogens.

        Returns
        -------
        all_combinations: list of tuples of ints
            e.g. [(1, 5, 9)]
        """
        comb = list(itertools.combinations(self.heavy_ids, 3))
        comb += list(itertools.combinations(self.h_ids, 3))

        combr = [x[::-1] for x in comb]
        all_comb = [x for items in zip(comb, combr) for x in items]
        return all_comb

    def add_orientations(self, orient=[], n_orient=0, translate=[],
                         n_translate=0, rotate=[], n_rotate=0,
                         load_files=False):
        """
        Add orientations to conformers.

        Parameters
        ----------
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given,
            ``n_orient`` orientations are automatically generated for the
            molcule. Heavy atoms are prioritised.
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given,
            ``n_rotate`` rotations are automatically generated for the
            molecule. Heavy atoms are prioritised.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given,
            ``n_translate`` translations are randomly generated for the
            molcule in the domain [0, 1).
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.
        """
        if not orient and n_orient:
            orient = self._orient_combs[:n_orient]

        if not rotate and n_rotate:
            rotate = self._orient_combs[:n_rotate]

        if not translate and n_translate:
            translate = np.random.rand(n_translate, 3)

        for conf in self.conformers:
            conf.add_orientations(orient=orient, rotate=rotate,
                                  translate=translate, load_files=load_files)

    def run(self, stage_2=True, opt=False, save_opt_geometry=False,
            chrconstr=[], chrequiv=[], basis='6-31g*', method='scf',
            psi4_options={}, hyp_a1=0.0005, hyp_a2=0.001, n_orient=0, orient=[],
            n_rotate=0, rotate=[], n_translate=0, translate=[],
            equal_methyls=False, restraint=True, load_files=False,
            **kwargs):
        """
        Perform a 1- or 2-stage ESP or RESP fit.

        Parameters
        ----------
        stage_2: bool (optional)
            Whether to perform a 2-stage RESP fit
        opt: bool (optional)
            Whether to optimise the geometry of each conformer
        save_opt_geometry: bool (optional)
            if ``True``, writes optimised geometries to an XYZ file
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        chrequiv: list (optional)
            lists of atoms with equivalent charges, indexed from 1.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.
        basis: str (optional)
            Basis set for QM
        method: str (optional)
            Method for QM
        psi4_options: dict (optional)
            additional Psi4 options
        hyp_a1: float (optional)
            scale factor of asymptote limits of hyperbola for first stage.
        hyp_a2: float (optional)
            scale factor of asymptote limits of hyperbola for second stage.
        restraint: bool (optional)
            whether to perform a restrained fit
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given,
            ``n_orient`` orientations are automatically generated for the
            molcule. Heavy atoms are prioritised.
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given,
            ``n_rotate`` rotations are automatically generated for the
            molecule. Heavy atoms are prioritised.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given,
            ``n_translate`` translations are randomly generated for the
            molcule in the domain [0, 1).
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent. 
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.
        **kwargs:
            arguments passed to Resp.fit

        Returns
        -------
        charges: ndarray
        """

        if opt:
            self.optimize_geometry(method=method, basis=basis,
                                   psi4_options=psi4_options,
                                   save_opt_geometry=save_opt_geometry)

        self.add_orientations(n_orient=n_orient, orient=orient,
                              n_rotate=n_rotate, rotate=rotate,
                              n_translate=n_translate,
                              translate=translate, load_files=load_files)

        if stage_2:  # do constraints in stage 2 only
            stage_2_equiv = chrequiv
            chrequiv = []
        else:
            stage_2_equiv = []

            if equal_methyls:
                methyls = self.get_methyl_constraints(chrconstr)
                chrequiv = list(chrequiv) + methyls

        q = self.fit(chrconstr=chrconstr,
                     chrequiv=chrequiv,
                     basis=basis, method=method,
                     hyp_a=hyp_a1, restraint=restraint,
                     **kwargs)

        if stage_2:
            if restraint:
                self.stage_1_charges = [self.unrestrained_charges,
                                        self.restrained_charges]
            else:
                self.stage_1_charges = self.unrestrained_charges
            cs = self.get_stage2_constraints(q, chrequiv=stage_2_equiv,
                                             equal_methyls=equal_methyls)
            constr, equiv = cs

            q = self.fit(chrconstr=constr, chrequiv=equiv,
                         basis=basis, method=method,
                         hyp_a=hyp_a2, restraint=restraint,
                         **kwargs)
        return q



import psiresp
import pytest
import os
import numpy as np

from numpy.testing import (assert_almost_equal, assert_equal,
                           assert_allclose)
from .utils import (mol_from_file, esp_from_gamess_file,
                    coordinates_from_xyz)


class BaseTestOrientation:

    molname = None
    orientname = None

    @pytest.fixture(scope='class')
    def opt_mol(self):
        return mol_from_file('{}.xyz'.format(self.orientname))

    @pytest.fixture(scope='function')
    def opt_orientation(self, opt_mol):
        return psiresp.Orientation(opt_mol, load_files=False)

    @pytest.fixture(scope='function')
    def esp(self):
        return esp_from_gamess_file('{}.esp'.format(self.orientname))

    @pytest.fixture(scope='function')
    def coordinates(self):
        return coordinates_from_xyz('{}.xyz'.format(self.orientname))

    def test_correct_coordinates(self, opt_orientation, coordinates):
        assert_almost_equal(opt_orientation.coordinates, coordinates)

    def test_get_grid(self, opt_orientation, esp):
        ref = esp[:, 1:]
        grid = opt_orientation.get_grid()
        assert_almost_equal(grid, ref, decimal=4)

    def test_get_esp_gas(self, opt_orientation, esp, tmpdir):
        ref = esp[:, 0]
        with tmpdir.as_cwd():
            opt_orientation.get_esp_matrices()
        epot = opt_orientation.esp
        assert_almost_equal(epot, ref, decimal=4)


@pytest.mark.fast
class TestOrientationDMSO0(BaseTestOrientation):
    orientname = 'dmso_opt_c1'


@pytest.mark.fast
class TestOrientationDMSO1(BaseTestOrientation):
    molname = 'dmso_c1'
    orientname = 'dmso_opt_c1_o1'

    @pytest.fixture(scope='class')
    def geometry(self):
        return mol_from_file('{}.xyz'.format(self.molname))

    @pytest.fixture(scope='function')
    def orientation(self, geometry):
        return psiresp.Orientation(geometry, load_files=False)

    def test_init_orientation(self, orientation):
        assert orientation.name == 'default'
        assert orientation.n_atoms == 10
        assert_equal(orientation.indices, np.arange(10).astype(int))
        assert_equal(orientation.symbols, list('CHHHSOCHHH'))
        assert orientation.grid is None
        assert orientation.esp is None
        assert orientation.r_inv is None


@pytest.mark.fast
class TestOrientationDMSO2(TestOrientationDMSO1):
    orientname = 'dmso_opt_c1_o2'



"""
psiresp
A RESP plugin for Psi4
"""

# Add imports here
from .orientation import Orientation
from .conformer import Conformer
from .resp import Resp
from .multiresp import MultiResp
from .resp2 import Resp2, MultiResp2
from .configs import (RespA1, RespA2, EspA1, EspA2, ATBResp,
                      MultiRespA1, MultiRespA2, MultiEspA1, MultiEspA2)
# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



from __future__ import division, absolute_import
import warnings
import itertools

import numpy as np

from . import utils


class MultiResp(object):
    """
    Class to manage R/ESP for multiple molecules of multiple conformers.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """

    def __init__(self, resps):
        self.molecules = resps
        self._unrestrained_charges = None
        self._restrained_charges = None

    def clone(self, suffix='_copy'):
        """Clone into another instance of MultiResp

        Parameters
        ----------
        suffix: str (optional)
            This is appended to each of the names of the molecule Resps
            in the MultiResp

        Returns
        -------
        MultiResp
        """
        names = [r.name+suffix for r in self.molecules]
        resps = [m.clone(name=n) for n, m in zip(names, self.molecules)]
        return type(self)(resps)

    @property
    def n_structures(self):
        return sum([mol.n_structures for mol in self.molecules])

    @property
    def n_molecules(self):
        return len(self.molecules)

    @property
    def restrained_charges(self):
        return self._restrained_charges

    @restrained_charges.setter
    def restrained_charges(self, charges):
        for mol, q in zip(self.molecules, charges):
            mol.restrained_charges = q
        self._restrained_charges = charges

    @property
    def unrestrained_charges(self):
        return self._unrestrained_charges

    @unrestrained_charges.setter
    def unrestrained_charges(self, charges):
        for mol, q in zip(self.molecules, charges):
            mol.unrestrained_charges = q
        self._unrestrained_charges = charges

    def _nmol_values(self, values, name, element):
        """
        Validate given values to check they are either a single value 
        or the same length as ``n_molecules``. Raises an error message 
        with the ``name`` of the variable and ``element`` description 
        if the lengths mismatch.
        """
        values = utils.iter_single(values)
        try:
            if len(values) != self.n_molecules:
                err = ('{} must be a list with the '
                       'same length as the number of molecules, '
                       'where each element is a {}')
                raise ValueError(err.format(name, element))
        except TypeError:  # itertools repeat
            pass
        return values

    def optimize_geometry(self, method='scf', basis='6-31g*',
                          psi4_options={}, save_opt_geometry=False,
                          save_files=False):
        """
        Optimise geometry for all molecules.

        Parameters
        ----------
        basis: str (optional)
            Basis set to optimise geometry
        method: str (optional)
            Method to optimise geometry
        psi4_options: dict (optional)
            additional Psi4 options
        save_opt_geometry: bool (optional)
            if ``True``, saves optimised geometries to XYZ file
        save_files: bool (optional)
            if ``True``, Psi4 files are saved. This does not affect 
            writing optimised geometries to files. 
        """
        for mol in self.molecules:
            mol.optimize_geometry(method=method, basis=basis,
                                  psi4_options=psi4_options,
                                  save_opt_geometry=save_opt_geometry,
                                  save_files=save_files)

    def get_constraint_matrices(self, intra_chrconstr=[], intra_chrequiv=[],
                                inter_chrconstr=[], inter_chrequiv=[],
                                weights=1, **kwargs):
        """
        Get A and B matrices to solve for charges, including charge constraints.

        Parameters
        ----------
        intra_chrconstr: list of lists or dicts (optional)
            Intramolecular charge constraints for each molecule in the 
            form of, where the list of constraints has the form 
            [{charge: atom_number_list}] or [[(charge, atom_number_list)]]. 
            The numbers are indexed from 1. 
            e.g. [{0: [1, 2]}, {0.2: [3, 4, 5]}]
            mean that atoms 1 and 2 together have a charge of 0 in the first
            molecule, whereas atoms 3, 4, and 5 combine to a charge of 0.2 in 
            the second molecule.
        intra_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges within each molecule. e.g. ::

                [
                  [[1, 2], [3, 4, 5]],
                  [[1, 3, 5, 7]]
                 ] 

            mean that atoms 1 and 2 in the first molecule have equal 
            charges; atoms 3, 4, and 5 in the first molecule have 
            equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
            charges.
        inter_chrconstr: list of lists (optional)
            Intermolecular charge constraints in the form of 
            {charge: [(mol_number, atom_number)]} or 
            [(charge, [(mol_number, atom_number)])]. 
            The numbers are indexed from 1.
            e.g. {0: [(1, 3), (2, 1)]} or [(0, [(1, 3), (2, 1)])] mean that
            the third atom of the first molecule, and the first atom of the 
            second molecule, combine to have a charge of 0.
        inter_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges between each molecule, 
            in the form [(mol_number, atom_number)]. 
            e.g. [[(1, 2), (2, 2), (3, 4)]] mean that atom 2 of molecule 1, 
            atom 2 of molecule 2, and atom 4 of molecule 3, all have equal 
            charges.
        weights: iterable (optional)
            List of weights for each molecule, or list of lists of weights for
            each conformer. e.g. [1, 2, [3, 4]] weights all conformers in the 
            first molecule by 1; all the conformers in the second molecule by 2;
            the first conformer in the third molecule by 3; and the second 
            conformer in the third molecule by 4. If only one value is given, it
            is repeated for every molecule.
        **kwargs
            Arguments passed to Resp.get_constraint_matrices.

        Returns
        -------
        A: ndarray
        B: ndarray
        edges: list of tuples
            list of start, end indices of the molecule atoms in `a` and `b`.
            For example, ``i, j = edges[3]`` such that ``a[i:j, i:j]`` is 
            the A matrix for the 4th molecule.
        nmol: ndarray
            array containing the number of orientations in each Resp, with 
            the same shape as B
        """
        intra_chrconstr = self._nmol_values(intra_chrconstr,
                                            'intra_chrconstr',
                                            'list or dict of constraints')
        intra_chrequiv = self._nmol_values(intra_chrequiv,
                                           'intra_chrequiv',
                                           'list of equivalence constraints')

        if utils.empty(inter_chrconstr):
            inter_chrconstr = []
        if utils.empty(inter_chrequiv):
            inter_chrequiv = []

        if not utils.isiterable(weights):
            weights = itertools.repeat(weights)
        elif len(weights) != self.n_molecules:
            err = ('weights must be an iterable of the same length '
                   'as number of molecules')
            raise ValueError(err)

        # gather individual A and B matrices for each Resp class
        a_s, b_s, nmol = [], [], []
        for m, constr, equiv, w in zip(self.molecules, intra_chrconstr,
                                       intra_chrequiv, weights):
            a, b = m.get_constraint_matrices(chrconstr=constr, chrequiv=equiv,
                                             weights=w, **kwargs)
            a_s.append(a)
            b_s.append(b)
            nmol.extend([m.n_structures]*len(b))

        mol_edges = np.r_[0, np.cumsum([len(b) for b in b_s])].astype(int)
        edges = list(zip(mol_edges[:-1], mol_edges[1:]))
        n_intra = mol_edges[-1]

        # allow for inter constraints to be given either as (nmol, natom)
        # or (nmol, [n_atom])
        def groups_to_indices(groups):
            indices = []
            for a, b in groups:
                offset = mol_edges[a-1]
                if isinstance(b, (tuple, list)):
                    indices.extend([offset+x-1 for x in b])
                else:
                    indices.append(offset+x-1)
            return np.array(indices)

        # set up intermolecular constraints
        if isinstance(inter_chrconstr, dict):
            inter_chrconstr = list(inter_chrconstr.items())
        inter_constr = [(q, groups_to_indices(e)) for q, e in inter_chrconstr]
        n_constr = len(inter_constr)

        inter_equiv = [groups_to_indices(e) for e in inter_chrequiv]
        inter_equiv = [e for e in inter_equiv if len(e) >= 2]
        lengths = np.cumsum([len(x)-1 for x in inter_equiv])
        equiv_edges = np.r_[0, lengths].astype(int)
        n_equiv = equiv_edges[-1]

        ndim = n_intra+n_equiv+n_constr
        nmol.extend([self.n_structures]*(n_equiv+n_constr))

        A = np.zeros((ndim, ndim))
        B = np.zeros(ndim)

        for (i, j), a, b in zip(edges, a_s, b_s):
            A[i:j, i:j] = a
            B[i:j] = b

        for i, (q, ix) in enumerate(inter_constr, n_intra):
            B[i] = q
            A[i, ix] = A[ix, i] = 1

        for i, ix in enumerate(inter_equiv):
            x = np.arange(equiv_edges[i], equiv_edges[i+1])+n_intra+n_constr
            A[(x, ix[:-1])] = A[(ix[:-1], x)] = -1
            A[(x, ix[1:])] = A[(ix[1:], x)] = 1

        edges = [[i, i+m.n_atoms] for (i, _), m in zip(edges, self.molecules)]
        return A, B, np.array(edges), np.array(nmol)

    def fit(self, restraint=True, hyp_a=0.0005, hyp_b=0.1, ihfree=True,
            tol=1e-5, maxiter=50, **kwargs):
        """
        Perform the R/ESP fits.

        Parameters
        ----------
        restraint: bool (optional)
            whether to perform a restrained fit
        hyp_a: float (optional)
            scale factor of asymptote limits of hyperbola
        hyp_b: float (optional)
            tightness of hyperbola at its minimum
        ihfree: bool (optional)
            if `True`, exclude hydrogens from restraint
        tol: float (optional)
            threshold for convergence
        maxiter: int (optional)
            maximum number of iterations
        **kwargs:
            arguments passed to MultiResp.get_constraint_matrices

        Returns
        -------
        charges: list of ndarray
        """
        a, b, edges, nmol = self.get_constraint_matrices(**kwargs)
        q = np.linalg.solve(a, b)
        self.unrestrained_charges = [q[i:j] for i, j in edges]
        if restraint:
            q = self.iter_solve(q, a, b, edges, nmol, hyp_a=hyp_a, hyp_b=hyp_b,
                                ihfree=ihfree, tol=tol, maxiter=maxiter)
            self.restrained_charges = [q[i:j] for i, j in edges]
        self.charges = [q[i:j] for i, j in edges]
        for mol, q in zip(self.molecules, self.charges):
            mol.charges = q
        return self.charges

    def iter_solve(self, q, a, b, edges, nmol, hyp_a=0.0005, hyp_b=0.1,
                   ihfree=True, tol=1e-5, maxiter=50):
        """
        Fit the charges iteratively, as required for the hyperbola penalty 
        function.

        Parameters
        ----------
        q: ndarray
            partial atomic charges
        a: ndarray
            unrestrained matrix A
        b: ndarray
            matrix B
        edges: list of tuples
            list of start, end indices for the atoms in ``a`` and ``b``
        nmol: ndarray
            array containing the number of orientations in each Resp, with 
            the same shape as ``b``
        hyp_a: float (optional)
            scale factor of asymptote limits of hyperbola
        hyp_b: float (optional)
            tightness of hyperbola at its minimum
        ihfree: bool (optional)
            if `True`, exclude hydrogens from restraint
        tol: float (optional)
            threshold for convergence
        maxiter: int (optional)
            maximum number of iterations

        Returns
        -------
        charges: ndarray
        """
        mask = np.zeros(len(b), dtype=bool)
        for i, j in edges:
            mask[i:j] = True
        if ihfree:
            for mol, (i, j) in zip(self.molecules, edges):
                mask[i:j] = mol.symbols != 'H'
        diag = np.diag_indices(len(b))
        ix = (diag[0][mask], diag[1][mask])
        b2 = hyp_b**2
        n_mol = nmol[mask]

        niter, delta = 0, 2*tol
        while delta > tol and niter < maxiter:
            q_last, a_i = q.copy(), a.copy()
            a_i[ix] = a[ix] + hyp_a/np.sqrt(q[mask]**2 + b2) * n_mol
            q = np.linalg.solve(a_i, b)
            delta = np.max((q-q_last)[mask]**2) ** 0.5
            niter += 1

        if delta > tol:
            err = 'Charge fitting did not converge with maxiter={}'
            warnings.warn(err.format(maxiter))

        return q

    def get_stage2_constraints(self, qs, equal_methyls=False,
                               intra_chrequiv=[], inter_chrconstr=[],
                               intra_chrconstr=[]):
        """
        Create constraints for Resp stage 2. Atom numbers are indexed from 1.

        Parameters
        ----------
        qs: ndarray
            list of charges for each molecule from stage 1
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent. Ignored if ``chrequiv`` constraints are provided.
        intra_chrconstr: list of lists or dicts (optional)
            Intramolecular charge constraints for each molecule in the 
            form of {charge: atom_number_list} or [(charge, atom_number_list)]. 
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]] 
            mean that atoms 1 and 2 together have a charge of 0.
        intra_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges within each molecule. e.g. ::

                [
                  [[1, 2], [3, 4, 5]],
                  [[1, 3, 5, 7]]
                 ] 

            mean that atoms 1 and 2 in the first molecule have equal 
            charges; atoms 3, 4, and 5 in the first molecule have 
            equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
            charges.
        inter_chrequiv: list of lists (optional)
            Original intermolecular charge equivalence constraints from 
            stage 1

        Returns
        -------
        intra_chrconstr: list
            Intramolecular charge constraints for each molecule in the 
            form of [[[charge, atom_number_list]]]
        intra_chrequiv: list
            Intramolecular charge equivalence constraints for each 
            molecule in the form of [[[atom_number_list]]]
        """

        intra_chrconstr = self._nmol_values(intra_chrconstr,
                                            'intra_chrconstr',
                                            'list or dict of constraints')
        intra_chrequiv = self._nmol_values(intra_chrequiv,
                                           'intra_chrequiv',
                                           'list of equivalence constraints')

        iconstr = []
        iequiv = []

        # add inter constraints too
        if inter_chrconstr is None:
            inter_chrconstr = []
        elif isinstance(inter_chrconstr, dict):
            inter_chrconstr = list(inter_chrconstr.items())

        # anything involved in a charge constraint cannot fluctuate freely
        chrconstr = [list() for i in range(self.n_molecules)]
        for _, groups in inter_chrconstr:
            for molid, atoms in groups:
                if isinstance(atoms, (tuple, list)):
                    chrconstr[molid-1].extend(atoms)
                else:
                    chrconstr[molid-1].append(atoms)

        for i, atoms in enumerate(intra_chrconstr):
            chrconstr[i].extend(atoms)  # avoid any tuple+list hijinks

        rows = zip(self.molecules, qs, intra_chrequiv, chrconstr)
        for i, (mol, q, eq, cr) in enumerate(rows):
            c, e = mol.get_stage2_constraints(q, chrequiv=eq, chrconstr=cr,
                                              equal_methyls=equal_methyls)
            iconstr.append(c)
            iequiv.append(e)

        return iconstr, iequiv

    def get_methyl_constraints(self):
        """
        Get charge equivalence arrays when all methyls are treated as 
        equivalent, and all methylenes are equivalent. Toggle this with 
        ``equal_methyls=True`` in ``run()``.

        Returns
        -------
        equivalence arrays: list of lists of ints
            List of equivalence arrays. First array contains methyl carbons;
            second contains methylene carbons; third contains methyl hydrogens;
            last contains methylene hydrogens.
        """
        equivs = [[], [], [], []]
        for mol in self.molecules:
            meqs = mol.get_methyl_constraints()
            for j, eqs in enumerate(meqs):
                equivs[j].append(eqs)
        return equivs

    def add_orientations(self, orient=[], n_orient=0, rotate=[],
                         n_rotate=0, translate=[], n_translate=0,
                         load_files=False):
        """
        Add orientations to each conformer of each Resp molecule. 

        Parameters
        ----------
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given, 
            ``n_orient`` orientations are automatically generated for each 
            molcule. Heavy atoms are prioritised.
        orient: list of lists (optional)
            List of lists of reorientations. 
            Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given, 
            ``n_rotate`` rotations are automatically generated for each 
            molecule. Heavy atoms are prioritised.
        rotate: list of lists (optional)
            List of lists of rotations. 
            Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given, 
            ``n_translate`` translations are randomly generated for each 
            molcule in the domain [0, 1).
        translate: list of lists (optional)
            List of lists of translations. 
            Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
            x coordinates, 0 to the y coordinates, and -0.5 to the z 
            coordinates.
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.
        """

        orient = utils.iter_single(orient)
        rotate = utils.iter_single(rotate)
        translate = utils.iter_single(translate)
        for mol, o, r, t in zip(self.molecules, orient, rotate, translate):
            mol.add_orientations(orient=o, n_orient=0, rotate=r,
                                 n_rotate=0, translate=t, n_translate=0,
                                 load_files=load_files)

    def run(self, stage_2=True, opt=False, save_opt_geometry=False,
            intra_chrconstr=[], intra_chrequiv=[], inter_chrconstr=[],
            inter_chrequiv=[], n_orient=0, orient=[], n_rotate=0, rotate=[],
            n_translate=0, translate=[], equal_methyls=False, basis='6-31g*',
            method='scf', psi4_options={}, hyp_a1=0.0005, hyp_a2=0.001,
            load_files=False, **kwargs):
        """
        Parameters
        ----------
        stage_2: bool (optional)
            Whether to perform a 2-stage RESP fit
        opt: bool (optional)
            Whether to optimise the geometry of each conformer
        save_opt_geometry: bool (optional)
            if ``True``, writes optimised geometries to an XYZ file
        intra_chrconstr: list of lists or dicts (optional)
            Intramolecular charge constraints for each molecule in the 
            form of {charge: atom_number_list} or [(charge, atom_number_list)]. 
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]] 
            mean that atoms 1 and 2 together have a charge of 0.
        intra_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges within each molecule. e.g. ::

                [
                  [[1, 2], [3, 4, 5]],
                  [[1, 3, 5, 7]]
                 ] 

            mean that atoms 1 and 2 in the first molecule have equal 
            charges; atoms 3, 4, and 5 in the first molecule have 
            equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
            charges.
        inter_chrconstr: list of lists (optional)
            Intermolecular charge constraints in the form of 
            {charge: [(mol_number, atom_number)]} or 
            [(charge, [(mol_number, atom_number)])]. 
            The numbers are indexed from 1.
            e.g. {0: [(1, 3), (2, 1)]} or [(0, [(1, 3), (2, 1)])] mean that
            the third atom of the first molecule, and the first atom of the 
            second molecule, combine to have a charge of 0.
        inter_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges between each molecule, 
            in the form [(mol_number, atom_number)]. 
            e.g. [[(1, 2), (2, 2), (3, 4)]] mean that atom 2 of molecule 1, 
            atom 2 of molecule 2, and atom 4 of molecule 3, all have equal 
            charges.
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given, 
            ``n_orient`` orientations are automatically generated for each 
            molcule. Heavy atoms are prioritised.
        orient: list of lists (optional)
            List of lists of reorientations. 
            Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given, 
            ``n_rotate`` rotations are automatically generated for each 
            molecule. Heavy atoms are prioritised.
        rotate: list of lists (optional)
            List of lists of rotations. 
            Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given, 
            ``n_translate`` translations are randomly generated for each 
            molcule in the domain [0, 1).
        translate: list of lists (optional)
            List of lists of translations. 
            Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
            x coordinates, 0 to the y coordinates, and -0.5 to the z 
            coordinates.
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent.
        basis: str (optional)
            Basis set for QM
        method: str (optional)
            Method for QM
        psi4_options: dict (optional)
            additional Psi4 options
        hyp_a1: float (optional)
            scale factor of asymptote limits of hyperbola for first stage.
        hyp_a2: float (optional)
            scale factor of asymptote limits of hyperbola for second stage.
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.
        **kwargs:
            arguments passed to MultiResp.fit

        Returns
        -------
        charges: list of ndarrays
        """

        if opt:
            for mol in self.molecules:
                mol.optimize_geometry(method=method, basis=basis,
                                      psi4_options=psi4_options,
                                      save_opt_geometry=save_opt_geometry)

        self.add_orientations(n_orient=n_orient, orient=orient,
                              n_rotate=n_rotate, rotate=rotate,
                              n_translate=n_translate,
                              translate=translate, load_files=load_files)

        intra_chrconstr = self._nmol_values(intra_chrconstr,
                                            'intra_chrconstr',
                                            'list or dict of constraints')
        intra_chrequiv = self._nmol_values(intra_chrequiv,
                                           'intra_chrequiv',
                                           'list of equivalence constraints')

        if stage_2:  # do intra-constraints in stage 2 only
            stage_2_equiv = intra_chrequiv
            intra_chrequiv = [[] for i in range(self.n_molecules)]
        else:
            stage_2_equiv = [[] for i in range(self.n_molecules)]

        if equal_methyls and not stage_2:
            equivs = self.get_methyl_constraints()
            intra_chrequiv = [x+y for x, y in zip(intra_chrequiv, equivs)]

        qs = self.fit(intra_chrconstr=intra_chrconstr,
                      intra_chrequiv=intra_chrequiv,
                      inter_chrconstr=inter_chrconstr,
                      inter_chrequiv=inter_chrequiv,
                      basis=basis, method=method, hyp_a=hyp_a1,
                      **kwargs)
        if stage_2:
            cs = self.get_stage2_constraints(qs, equal_methyls=equal_methyls,
                                             intra_chrequiv=stage_2_equiv, 
                                             inter_chrconstr=inter_chrconstr,
                                             intra_chrconstr=intra_chrconstr)
            intra_c, intra_e = cs
            qs = self.fit(intra_chrconstr=intra_c,
                          intra_chrequiv=intra_e,
                          basis=basis, method=method,
                          hyp_a=hyp_a2, **kwargs)
        return qs



import numpy as np

from .conformer import Conformer
from .resp import Resp
from .multiresp import MultiResp
from . import utils
from .due import due, Doi


@due.dcite(
    Doi('10.1038/s42004-020-0291-4'),
    description='RESP2',
    path='psiresp.resp2',
)
class Resp2(object):
    """
    Class to manage Resp2 for one molecule of multiple conformers.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    gas: Resp
        Resp class of molecule in gaseous phase
    solv: Resp
        Resp class of molecule in aqueous phase
    gas_charges: ndarray of floats
        RESP charges in gas phase (only exists after calling run)
    solv_charges: ndarray of floats
        RESP charges in aqueous phase (only exists after calling run)
    charges: ndarray of floats
        Resp2 charges (only exists after calling run)
    """

    @classmethod
    def from_molecules(cls, molecules, charge=0, multiplicity=1, name=None,
                       orient=[], rotate=[], translate=[], grid_name='grid.dat',
                       esp_name='grid_esp.dat', load_files=False, **kwargs):
        """
        Create Resp class from Psi4 molecules.

        Parameters
        ----------
        molecules: iterable of Psi4 molecules
            conformers of the molecule. They must all have the same atoms
            in the same order.
        charge: int (optional)
            overall charge of the molecule.
        multiplicity: int (optional)
            multiplicity of the molecule
        name: str (optional)
            name of the molecule. This is used to name output files. If not
            given, defaults to 'Mol'.
        orient: list of tuples of ints (optional)
            List of reorientations to add to each conformer.
            Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        rotate: list of tuples of ints (optional)
            List of rotations to add to each conformer.
            Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        translate: list of tuples of floats (optional)
            List of translations to add to each conformer.
            Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        grid_name: str (optional)
            template for grid filename for each Orientation.
        esp_name: str (optional)
            template for ESP filename for each Orientation.
        load_files: bool (optional)
            If ``True``, tries to load data from file for each Orientation.
        **kwargs:
            arguments passed to ``Resp2.__init__()``.

        Returns
        -------
        resp2: Resp2
        """
        molecules = utils.asiterable(molecules)
        if name is not None:
            names = ['{}_c{}'.format(name, i+1) for i in range(len(molecules))]
        else:
            molnames = [m.name() for m in molecules]
            gennames = ['Mol_c{}'.format(i+1) for i in range(len(molecules))]
            names = np.where(molnames == 'default', molnames, gennames)

        conformers = []
        for mol, name in zip(molecules, names):
            conformers.append(Conformer(mol.clone(), name=name, charge=charge,
                                        multiplicity=multiplicity,
                                        orient=orient, rotate=rotate,
                                        translate=translate,
                                        grid_name=grid_name,
                                        esp_name=esp_name,
                                        load_files=load_files))

        return cls(conformers, name=name, **kwargs)

    def __init__(self, conformers, name=None, chrconstr=[],
                 chrequiv=[]):
        if name is None:
            name = 'Resp'
        self.name = name
        self.gas = Resp(conformers).clone(name=name+'_gas')
        self.solv = self.gas.clone(name=name+'_solv')
        self._gas_charges = self._solv_charges = self._charges = None
        if chrconstr is not None:
            self.gas.add_charge_constraints(chrconstr)
        if chrequiv is not None:
            self.solv.add_charge_equivalences(chrequiv)

    @property
    def gas_charges(self):
        if self._gas_charges is None:
            raise ValueError('No gas charges available. Call run()')
        return self._gas_charges

    @property
    def solv_charges(self):
        if self._solv_charges is None:
            raise ValueError('No solv charges available. Call run()')
        return self._solv_charges

    @property
    def charges(self):
        if self._charges is None:
            raise ValueError('No Resp2 charges available. Call run()')
        return self._charges

    def run(self, opt=False, save_opt_geometry=False, delta=0.6,
            chrconstr=[], chrequiv=[], weights=1, method='PW6B95',
            basis='aug-cc-pV(D+d)Z', vdw_radii={}, psi4_options={},
            rmin=1.3, rmax=2.1, save_files=False, n_orient=0, orient=[],
            n_rotate=0, rotate=[], n_translate=0, translate=[],
            equal_methyls=False, tol=1e-6, maxiter=50, load_files=False):
        """
        Perform a 2-stage RESP2 fit.

        Parameters
        ----------
        opt: bool (optional)
            Whether to optimise the geometry of each conformer
        save_opt_geometry: bool (optional)
            if ``True`` , writes optimised geometries to an XYZ file
        delta: float (optional)
            mixing parameter for aqueous and gaseous charges. delta=0.6 
            generates charges from 60% aqueous and 40% gaseous charges.
        chrconstr: list or dict (optional)
            Intramolecular charge constraints in the form of
            {charge: atom_number_list} or [(charge, atom_number_list)].
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
            mean that atoms 1 and 2 together have a charge of 0.
        chrequiv: list (optional)
            lists of atoms with equivalent charges.
            e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
            charges, and atoms 3, 4, and 5 have equal charges.
        weights: iterable (optional)
            weights of each conformer. If only one number is given, this is 
            repeated for each conformer.
        method: str (optional)
            Method to compute ESP
        basis: str (optional)
            Basis set to compute ESP
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
        psi4_options: dict (optional)
            additional Psi4 options
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed ESP
            and grids are written to files.
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given,
            ``n_orient`` orientations are automatically generated for the
            molcule. Heavy atoms are prioritised.
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given,
            ``n_rotate`` rotations are automatically generated for the
            molecule. Heavy atoms are prioritised.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given,
            ``n_translate`` translations are randomly generated for the
            molcule in the domain [0, 1).
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent. Ignored if ``chrequiv`` constraints are provided.
        tol: float (optional)
            threshold for fitting convergence
        maxiter: int (optional)
            maximum number of iterations in fitting
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.

        Returns
        -------
        charges: ndarray
        """
        if opt:
            for m, b in [('hf', '6-31g*'),
                         ('hf', basis),
                         (method, basis)]:
                self.gas.optimize_geometry(method=m, basis=b,
                                           psi4_options=psi4_options,
                                           save_opt_geometry=save_opt_geometry)

            self.solv = self.gas.clone(name=self.name+'_solv')

        self.gas.add_orientations(n_orient=n_orient, orient=orient,
                                  n_rotate=n_rotate, rotate=rotate,
                                  n_translate=n_translate,
                                  translate=translate)

        self.solv.add_orientations(n_orient=n_orient, orient=orient,
                                   n_rotate=n_rotate, rotate=rotate,
                                   n_translate=n_translate,
                                   translate=translate)

        self._gas_charges = self.gas.run(stage_2=True, opt=False,
                                         chrconstr=chrconstr,
                                         chrequiv=chrequiv,
                                         weights=weights, use_radii='bondi',
                                         vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                         vdw_point_density=2.5,
                                         vdw_radii=vdw_radii,
                                         rmin=rmin, rmax=rmax, method=method,
                                         basis=basis, solvent=None,
                                         restraint=True,
                                         psi4_options=psi4_options,
                                         hyp_a1=0.0005, hyp_a2=0.001,
                                         equal_methyls=equal_methyls,
                                         ihfree=True, tol=tol, maxiter=maxiter,
                                         save_files=save_files,
                                         load_files=load_files)

        self._solv_charges = self.solv.run(stage_2=True, opt=False,
                                           chrconstr=chrconstr,
                                           chrequiv=chrequiv,
                                           weights=weights, use_radii='bondi',
                                           vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                           vdw_point_density=2.5,
                                           vdw_radii=vdw_radii,
                                           rmin=rmin, rmax=rmax,
                                           method=method,
                                           basis=basis, solvent='water',
                                           restraint=True,
                                           psi4_options=psi4_options,
                                           hyp_a1=0.0005, hyp_a2=0.001,
                                           equal_methyls=equal_methyls,
                                           ihfree=True, tol=tol, maxiter=maxiter,
                                           save_files=save_files,
                                           load_files=load_files)

        self._charges = delta*self.solv_charges + (1-delta)*self.gas_charges
        return self._charges


@due.dcite(
    Doi('10.1038/s42004-020-0291-4'),
    description='RESP2 multi-molecule fit',
    path='psiresp.resp2',
)
class MultiResp2(object):
    """
    Class to manage Resp2 for multiple molecules of multiple conformers.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    gas: Resp
        MultiResp class of molecules in gaseous phase
    solv: Resp
        MultiResp class of molecules in aqueous phase
    gas_charges: ndarray of floats
        RESP charges in gas phase (only exists after calling run)
    solv_charges: ndarray of floats
        RESP charges in aqueous phase (only exists after calling run)
    charges: ndarray of floats
        Resp2 charges (only exists after calling run)
    """

    def __init__(self, resps):
        cresps = [r.clone(name=r.name+'_gas') for r in resps]
        self.gas = MultiResp(cresps)
        self.solv = self.gas.clone(suffix='_solv')
        self._gas_charges = self._solv_charges = self._charges = None

    @property
    def gas_charges(self):
        if self._gas_charges is None:
            raise ValueError('No gas charges available. Call run()')
        return self._gas_charges

    @property
    def solv_charges(self):
        if self._solv_charges is None:
            raise ValueError('No solv charges available. Call run()')
        return self._solv_charges

    @property
    def charges(self):
        if self._charges is None:
            raise ValueError('No Resp2 charges available. Call run()')
        return self._charges

    def run(self, opt=False, save_opt_geometry=False, delta=0.6,
            intra_chrconstr=[], intra_chrequiv=[], inter_chrconstr=[],
            inter_chrequiv=[], weights=1, method='PW6B95',
            basis='aug-cc-pV(D+d)Z', vdw_radii={}, psi4_options={},
            rmin=1.3, rmax=2.1, save_files=False, n_orient=0, orient=[],
            n_rotate=0, rotate=[], n_translate=0, translate=[],
            equal_methyls=False, tol=1e-6, maxiter=50, load_files=False):
        """
        Perform a 2-stage RESP2 fit.

        Parameters
        ----------
        opt: bool (optional)
            Whether to optimise the geometry of each conformer
        save_opt_geometry: bool (optional)
            if ``True``, writes optimised geometries to an XYZ file
        delta: float (optional)
            mixing parameter for aqueous and gaseous charges. delta=0.6 
            generates charges from 60% aqueous and 40% gaseous charges.
        intra_chrconstr: list of lists or dicts (optional)
            Intramolecular charge constraints for each molecule in the 
            form of {charge: atom_number_list} or [(charge, atom_number_list)]. 
            The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]] 
            mean that atoms 1 and 2 together have a charge of 0.
        intra_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges within each molecule. e.g. ::

                [
                  [[1, 2], [3, 4, 5]],
                  [[1, 3, 5, 7]]
                 ] 

            mean that atoms 1 and 2 in the first molecule have equal 
            charges; atoms 3, 4, and 5 in the first molecule have 
            equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
            charges.
        inter_chrconstr: list of lists (optional)
            Intermolecular charge constraints in the form of 
            {charge: [(mol_number, atom_number)]} or 
            [(charge, [(mol_number, atom_number)])]. 
            The numbers are indexed from 1.
            e.g. {0: [(1, 3), (2, 1)]} or [(0, [(1, 3), (2, 1)])] mean that
            the third atom of the first molecule, and the first atom of the 
            second molecule, combine to have a charge of 0.
        inter_chrequiv: list of lists (optional)
            Lists of atoms with equivalent charges between each molecule, 
            in the form [(mol_number, atom_number)]. 
            e.g. [[(1, 2), (2, 2), (3, 4)]] mean that atom 2 of molecule 1, 
            atom 2 of molecule 2, and atom 4 of molecule 3, all have equal 
            charges.
        weights: iterable (optional)
            weights of each conformer. If only one number is given, this is 
            repeated for each conformer.
        method: str (optional)
            Method to compute ESP
        basis: str (optional)
            Basis set to compute ESP
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
        psi4_options: dict (optional)
            additional Psi4 options
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        save_files: bool (optional)
            if ``True`` , Psi4 files are saved and the computed ESP
            and grids are written to files.
        n_orient: int (optional)
            If this is greater than 0 and ``orient`` is not given,
            ``n_orient`` orientations are automatically generated for the
            molcule. Heavy atoms are prioritised.
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_rotate: int (optional)
            If this is greater than 0 and ``rotate`` is not given,
            ``n_rotate`` rotations are automatically generated for the
            molecule. Heavy atoms are prioritised.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse
            order.
        n_translate: int (optional)
            If this is greater than 0 and ``translate`` is not given,
            ``n_translate`` translations are randomly generated for the
            molcule in the domain [0, 1).
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        equal_methyls: bool (optional)
            if ``True``, all carbons in methyl groups are constrained to be
            equivalent; all carbons in methylenes are equivalent; all hydrogens
            in methyls are equivalent; and all hydrogens in methylenes are
            equivalent. Ignored if ``chrequiv`` constraints are provided.
        tol: float (optional)
            threshold for fitting convergence
        maxiter: int (optional)
            maximum number of iterations in fitting
        load_files: bool (optional)
            If ``True``, tries to load ESP and grid data from file.

        Returns
        -------
        charges: ndarray
        """

        if opt:
            for m, b in [('hf', '6-31g*'),
                         ('hf', basis),
                         (method, basis)]:
                self.gas.optimize_geometry(method=m, basis=b,
                                           psi4_options=psi4_options,
                                           save_opt_geometry=save_opt_geometry)

            self.solv = self.gas.clone(suffix='_solv')

        self.add_orientations(n_orient=n_orient, orient=orient,
                              n_rotate=n_rotate, rotate=rotate,
                              n_translate=n_translate, translate=translate)

        self._gas_charges = self.gas.run(stage_2=True, opt=False,
                                         intra_chrconstr=intra_chrconstr,
                                         intra_chrequiv=intra_chrequiv,
                                         inter_chrconstr=inter_chrconstr,
                                         inter_chrequiv=inter_chrequiv,
                                         weights=weights, use_radii='bondi',
                                         vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                         vdw_point_density=2.5, vdw_radii=vdw_radii,
                                         rmin=rmin, rmax=rmax, method=method,
                                         basis=basis, solvent=None, restraint=True,
                                         psi4_options=psi4_options, hyp_a1=0.0005,
                                         hyp_a2=0.001, equal_methyls=equal_methyls,
                                         ihfree=True, tol=tol, maxiter=maxiter,
                                         save_files=save_files,
                                         load_files=load_files)

        self._solv_charges = self.solv.run(stage_2=True, opt=False,
                                           intra_chrconstr=intra_chrconstr,
                                           intra_chrequiv=intra_chrequiv,
                                           inter_chrconstr=inter_chrconstr,
                                           inter_chrequiv=inter_chrequiv,
                                           weights=weights, use_radii='bondi',
                                           vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                           vdw_point_density=2.5, vdw_radii=vdw_radii,
                                           rmin=rmin, rmax=rmax, method=method,
                                           basis=basis, solvent='water', restraint=True,
                                           psi4_options=psi4_options, hyp_a1=0.0005,
                                           hyp_a2=0.001, equal_methyls=equal_methyls,
                                           ihfree=True, tol=tol, maxiter=maxiter,
                                           save_files=save_files,
                                           load_files=load_files)
        charges = []
        for solv, gas, sresp, gresp in zip(self.solv_charges, self.gas_charges,
                                           self.gas.molecules, self.solv.molecules):
            sresp.charges = gresp.charges = q = delta*solv + (1-delta)*gas
            charges.append(q)

        self._charges = charges
        return charges



import argparse
import os
import re
import glob
import shutil
import subprocess as sp
from tempfile import TemporaryDirectory
from contextlib import contextmanager
# YAML imports
try:
    import yaml  # PyYAML
    loader = yaml.load
except ImportError:
    try:
        import ruamel_yaml as yaml  # Ruamel YAML
    except ImportError:
        try:
            # Load Ruamel YAML from the base conda environment
            from importlib import util as import_util
            CONDA_BIN = os.path.dirname(os.environ['CONDA_EXE'])
            ruamel_yaml_path = glob.glob(os.path.join(CONDA_BIN, '..',
                                                      'lib', 'python*.*', 'site-packages',
                                                      'ruamel_yaml', '__init__.py'))[0]
            # Based on importlib example, but only needs to load_module since its the whole package, not just
            # a module
            spec = import_util.spec_from_file_location('ruamel_yaml', ruamel_yaml_path)
            yaml = spec.loader.load_module()
        except (KeyError, ImportError, IndexError):
            raise ImportError("No YAML parser could be found in this or the conda environment. "
                              "Could not find PyYAML or Ruamel YAML in the current environment, "
                              "AND could not find Ruamel YAML in the base conda environment through CONDA_EXE path. " 
                              "Environment not created!")
    loader = yaml.YAML(typ="safe").load  # typ="safe" avoids odd typing on output


@contextmanager
def temp_cd():
    """Temporary CD Helper"""
    cwd = os.getcwd()
    with TemporaryDirectory() as td:
        try:
            os.chdir(td)
            yield
        finally:
            os.chdir(cwd)


# Args
parser = argparse.ArgumentParser(description='Creates a conda environment from file for a given Python version.')
parser.add_argument('-n', '--name', type=str,
                    help='The name of the created Python environment')
parser.add_argument('-p', '--python', type=str,
                    help='The version of the created Python environment')
parser.add_argument('conda_file',
                    help='The file for the created Python environment')

args = parser.parse_args()

# Open the base file
with open(args.conda_file, "r") as handle:
    yaml_script = loader(handle.read())

python_replacement_string = "python {}*".format(args.python)

try:
    for dep_index, dep_value in enumerate(yaml_script['dependencies']):
        if re.match('python([ ><=*]+[0-9.*]*)?$', dep_value):  # Match explicitly 'python' and its formats
            yaml_script['dependencies'].pop(dep_index)
            break  # Making the assumption there is only one Python entry, also avoids need to enumerate in reverse
except (KeyError, TypeError):
    # Case of no dependencies key, or dependencies: None
    yaml_script['dependencies'] = []
finally:
    # Ensure the python version is added in. Even if the code does not need it, we assume the env does
    yaml_script['dependencies'].insert(0, python_replacement_string)

# Figure out conda path
if "CONDA_EXE" in os.environ:
    conda_path = os.environ["CONDA_EXE"]
else:
    conda_path = shutil.which("conda")
if conda_path is None:
    raise RuntimeError("Could not find a conda binary in CONDA_EXE variable or in executable search path")

print("CONDA ENV NAME  {}".format(args.name))
print("PYTHON VERSION  {}".format(args.python))
print("CONDA FILE NAME {}".format(args.conda_file))
print("CONDA PATH      {}".format(conda_path))

# Write to a temp directory which will always be cleaned up
with temp_cd():
    temp_file_name = "temp_script.yaml"
    with open(temp_file_name, 'w') as f:
        f.write(yaml.dump(yaml_script))
    sp.call("{} env create -n {} -f {}".format(conda_path, args.name, temp_file_name), shell=True)
